/* #undef SDL_VENDOR_INFO */
#define SDL_REVISION_NUMBER 0

#ifdef SDL_VENDOR_INFO
#define SDL_REVISION "SDL-prerelease-2.29.2-352-g10b4a7937 (" SDL_VENDOR_INFO ")"
#else
#define SDL_REVISION "SDL-prerelease-2.29.2-352-g10b4a7937"
#endif
