/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017 Nathan Osman
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>

#include <qmdnsengine/server.h>
#include <qmdnsengine/resolver.h>

class QItemSelection;
class QLineEdit;
class QListView;
class QListWidget;
class QPushButton;
class QTableWidget;

class ServiceModel;

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:

    MainWindow();

private Q_SLOTS:

    void onToggled(bool checked);
    void onClicked();
    void onSelectionChanged(const QItemSelection &selected, const QItemSelection &deselected);

private:

    QMdnsEngine::Server mServer;
    ServiceModel *mServiceModel;

    QLineEdit *mServiceType;
    QPushButton *mStartStop;
    QListView *mServices;
    QListWidget *mAddresses;
    QTableWidget *mAttributes;

    QMdnsEngine::Resolver *mResolver;
};

#endif // MAINWINDOW_H
